///////////////////////////////////////////////////////////////////////////////
//	
//								Nano OS	Version	1.0
//						Copyright 2009 xxxxxxx, xxxxxxx
//	File:	Nano_Queue.c
//	Author: Balasubramanya Bhat
//	Description: Main Scheduling function implementation
//	
///////////////////////////////////////////////////////////////////////////////

#include "NanoOS.h"
#include "Nano_Queue.h"
#include "TI_C6713\Nano_Config.h"

Nano_Queue g_ready_q;
Nano_Queue g_wait_q;
Nano_Queue g_ap_wait_q;

// This global variable can be accessed from outside
volatile BOOL OS_IsRunning = FALSE;

volatile UINT64 g_global_time;	// This variable gets updated everytime the Timer ISR is called.
volatile UINT64 g_next_wakeup_time; // This variable holds the next scheduled wakeup time in uSecs
volatile void * g_current_task;

OS_AperiodicTask g_TCB_idle_task;	// A TCB for the idle task

// External functions used in here
extern void Nano_InitTimer();
void Nano_Timer1ISRHook();
void Nano_Timer2ISRHook();
extern void C6713_ContextRestore(void *new_task);
extern void C6713_ContextSw(void * new_task);
void Nano_SetAlarm(OS_PeriodicTask *task, UINT64 abs_time_in_us, BOOL is_new_job);
void  Nano_UpdateTaskBudget (OS_PeriodicTask *task);

///////////////////////////////////////////////////////////////////////////////
// The following funcstion starts the OS scheduling
///////////////////////////////////////////////////////////////////////////////
void OS_Start()
{
	// Check if the OS is already running
	if(!OS_IsRunning)
	{
		OS_IsRunning = TRUE;	// No need to protect this line and the above
		// 'if' condition as this is called from the initialization for the
		// first time and there is only single thread at that point of time.

		// Reset the global timer
		g_global_time = 0;
		g_next_wakeup_time = 0;

		// Reset the current task
		g_current_task = 0;

		// Initialize the IDLE task TCB. This is done here so that the 
		g_TCB_idle_task.id = 0;
		g_TCB_idle_task.priority = MIN_PRIORITY + 1;
		g_TCB_idle_task.top_of_stack = 0;
		g_TCB_idle_task.type = APERIODIC;
		g_current_task = &g_TCB_idle_task;	// Set the current task as the idle task

		// Add the idle task to the Aperiodic task queue
		Nano_QueueInsert(&g_ap_wait_q, &g_TCB_idle_task, MIN_PRIORITY + 1);		

		// Start scheduling by creating the first interrupt
		Nano_UpdateTimer(1000);

		// Start processing interrupts
		Nano_InitInterrupts();

		// Now this task continues as the idle task
		while(1)
		{}
	}
}

///////////////////////////////////////////////////////////////////////////////
// The following funcstion starts triggers OS scheduling
///////////////////////////////////////////////////////////////////////////////
void Nano_ReSchedule()
{
	void * task;

	OS_ENTER_CRITICAL();	// Enter critical section

	if(Nano_QueuePeek(&g_ready_q, (void**) &task, 0) != SUCCESS)
	{
		Nano_QueuePeek(&g_ap_wait_q, (void**) &task, 0);
	}

	Nano_UpdateTaskBudget(task);

	OS_EXIT_CRITICAL();	// Exit critical section

	C6713_ContextSw(task);	// This has the affect of g_current_task = task;
}

///////////////////////////////////////////////////////////////////////////////
// This function is invoked when the timer expires. This function should calculate
// and return the new delay required for the next timer expiry in number of
// microseconds.
///////////////////////////////////////////////////////////////////////////////
void Nano_Timer1ISRHook()
{
	OS_PeriodicTask * task;
	UINT64 new_time = 0;
	
	g_global_time = g_next_wakeup_time;	
	g_next_wakeup_time = 0xFFFFFFFFFFFFFFFF;

	while(Nano_QueuePeek(&g_ready_q, (void**) &task, &new_time) == SUCCESS)
	{
		// First Check the deadline expiry for each job in the ready queue
		if(new_time == g_global_time)
		{
			// Now get the new task from the queue.
			Nano_QueueGet(&g_ready_q, (void**) &task, 0);

			// Deadline has expired
			// TODO: Take necessary action for deadline miss
			task->dline_miss_count ++;

			// If the relative deadline and the period are same, then reintroduce a job
			// in the same queue else put it in the wait_q
			if(task->deadline == task->period)
			{
				// ReSet the remaining budget to full budget
				task->remaining_budget = task->budget;
				Nano_SetAlarm(task, g_global_time + task->period, TRUE);
			}
			else
			{
				Nano_SetAlarm(task, g_global_time + task->period - task->deadline, FALSE);
			}

			continue;
		}
		else
		{
			break;				
		}
	}

	// Now consider new jobs to be introduced from the wait queue
	while(Nano_QueuePeek(&g_wait_q, (void**) &task, &new_time) == SUCCESS)
	{
		// First Check the deadline expiry for each job in the ready queue
		if(new_time == g_global_time)
		{
			// Now get the new task from the queue.
			Nano_QueueGet(&g_wait_q, (void**) &task, 0);
			
			// ReSet the remaining budget to full budget
			task->remaining_budget = task->budget;
			Nano_SetAlarm(task, g_global_time + task->deadline, TRUE);
			continue;
		}
		else
		{
			break;				
		}
	}

	if(Nano_QueuePeek(&g_ready_q, (void**) &task, 0) != SUCCESS)
	{
		Nano_QueuePeek(&g_ap_wait_q, (void**) &task, 0);
	}

	Nano_UpdateTaskBudget(task);
	C6713_ContextRestore(task);	// This has the affect of g_current_task = task;					
}

///////////////////////////////////////////////////////////////////////////////
// ISR for the Timer2 (Budget)Timer
// Some thread has exceded its budget. This function handles this case.
///////////////////////////////////////////////////////////////////////////////
void Nano_Timer2ISRHook()
{
	OS_PeriodicTask * task = (OS_PeriodicTask *)g_current_task;
	UINT32 timeout = 0;
	if(task->type == PERIODIC )
	{
		// The thread budget has exceeded
		task->TBE_count++;

		// TODO: Raise the flag saying that the thread has exceeded its
		// budget
		
		// Add to the thread's accumulated budget
		task->accumulated_budget += task->remaining_budget;
		task->remaining_budget = 0;

		// Suspend the current task. This task will be automatically 
		// woken up by the alarm.
		Nano_QueueDelete(&g_ready_q, task);
		if(task->alarm_time == g_next_wakeup_time)
		{
			// If the next wakeup time is same as the alarm time for the 
			// current task, just invalidate the next wakeup time so that
			// it is set again below
			g_next_wakeup_time = 0xFFFFFFFFFFFFFFFF;
		}
		if(task->deadline == task->period)
		{
			Nano_SetAlarm(task, task->alarm_time, FALSE);
		}
		else
		{
			Nano_SetAlarm(task, task->alarm_time + task->period - task->deadline, FALSE);
		}

		if(Nano_QueuePeek(&g_ready_q, (void**) &task, 0) != SUCCESS)
		{
			Nano_QueuePeek(&g_ap_wait_q, (void**) &task, 0);
		}

		// Set the budget timer
		timeout = (task->type == PERIODIC) ? task->remaining_budget : 0;
		Nano_SetBudgetTimer(timeout);			
		C6713_ContextRestore(task);	// This has the affect of g_current_task = task;					
	}
}

///////////////////////////////////////////////////////////////////////////////                                           TASK SWITCH HOOK
//
// Description: This function is called before a task switch is performed.  
// This allows you to perform other operations during a context switch.
//
// Arguments  : none
//
// Note(s)    : 1) Interrupts shall be disabled before making this function call
///////////////////////////////////////////////////////////////////////////////
void  Nano_UpdateTaskBudget (OS_PeriodicTask *task)
{
	OS_PeriodicTask * old_task = (OS_PeriodicTask *)g_current_task;
	UINT32 old_time = 0, new_timeout;

	// If the new and the old tasks are same, then just return.
	if(old_task == task) return;

	// Check if the new task is a valid periodic thread. 
	new_timeout = (task->type == PERIODIC) ? task->remaining_budget : 0;

	old_time = Nano_SetBudgetTimer(new_timeout);			

	// Check if the old task is a valid PERIODIC thread. 
	if(old_task->type == PERIODIC)
	{
		old_task->remaining_budget -= old_time;

		// Add to the thread's accumulated budget
		old_task->accumulated_budget += old_time;
	}
}

///////////////////////////////////////////////////////////////////////////////
// This function puts the current thread to sleep and sets up an alarm for
// waking up at a specified time instant (relative to the starting time).
// ASSUMPTION: The interrupts are disabled when this function is invoked
///////////////////////////////////////////////////////////////////////////////
void Nano_SetAlarm(OS_PeriodicTask *task, UINT64 abs_time_in_us, BOOL is_new_job)
{
	Nano_Queue * q;
	UINT64 t1 = 0xFFFFFFFFFFFFFFFF;
	UINT64 t2 = 0xFFFFFFFFFFFFFFFF;

	// If it was a past time, then return immediately.
	if(g_global_time >= abs_time_in_us) 
	{
		return;
	}

	// Insert the task into the g_ready_q / g_wait_q
	task->alarm_time = abs_time_in_us;
	q = (is_new_job) ? &g_ready_q : &g_wait_q;

	Nano_QueueInsert(q, (void *) task, abs_time_in_us);

	// Set the timer to the next earliest timeout requested in both queues
	Nano_QueuePeek(&g_ready_q, 0, &t1);
	Nano_QueuePeek(&g_wait_q, 0, &t2);
	abs_time_in_us = (t1 < t2) ? t1 : t2;
			
	if((abs_time_in_us > g_global_time) && \
		(abs_time_in_us < g_next_wakeup_time))	
	{
		// We want a shorter timeout.
		g_next_wakeup_time = abs_time_in_us;
		Nano_UpdateTimer((UINT32)(g_next_wakeup_time - g_global_time));		
	}
}
