///////////////////////////////////////////////////////////////////////////////
//	
//								Nano OS	Version	1.0
//						Copyright 2009 xxxxxxx, xxxxxxx
//	File:	C6713_Nano_Task.asm
//	Author:	Balasubramanya bhat (bbhat@ncsu.edu)
//	Description: OS Task Related routines
//	
///////////////////////////////////////////////////////////////////////////////

#include "NanoOS.h"
#include "Nano_Queue.h"

// function prototype declaration
static BOOL ValidateNewThread(UINT32 period, UINT32 budget);

#ifdef _USE_STD_LIBS
	#define FAULT printf
#else
	#define FAULT
#endif

// Macro for calculating the thread usage
#define CALC_THREAD_CPU_USAGE(period, budget) ((FP32)budget / (FP32)period)

// Variable declaration/definition
static UINT32 g_task_id_counter = 0;
static FP32 g_total_allocated_cpu = 0.0;

extern Nano_Queue g_ready_q;
extern Nano_Queue g_wait_q;
extern Nano_Queue g_ap_wait_q;
extern volatile UINT64 g_global_time;	// This variable gets updated everytime the Timer ISR is called.
extern volatile UINT64 g_next_wakeup_time; // This variable holds the next scheduled wakeup time in uSecs


extern UINT32 * C6713_BuildTaskStack(UINT32 * stack_ptr, void (*task_function)(void *), void * arg);
static void TaskEntryMain(void *pdata);
void Nano_ReSchedule();

///////////////////////////////////////////////////////////////////////////////
// OS_CreatePeriodicTask - API to create periodic tasks
///////////////////////////////////////////////////////////////////////////////
OS_Error OS_CreatePeriodicTask(
	UINT32 period_in_us,
	UINT32 deadline_in_us,
	UINT32 budget_in_us,
	UINT32 phase_shift_in_us,
	UINT32 *stack,
	UINT32 stack_size_in_bytes,
	OS_PeriodicTask *task,
	void (*periodic_entry_function)(void *pdata),
	void *pdata)
{
	FP32 this_thread_cpu;
	UINT32 stack_size;

	if(!task)
	{
	    FAULT("Invalid Task");
		return INVALID_TASK;
	}

    if(!periodic_entry_function || !stack)
	{
	    FAULT("One or more invalid %s arguments", "task");
		return INVALID_ARG;
	}

	if(stack_size_in_bytes < ONE_KB / 4)	// Validate for minimum stack size 256 bytes
	{
		FAULT("Stack size should be at least %d bytes", ONE_KB / 4);
		return INSUFFICIENT_STACK;
	}

	if(period_in_us < MIN_PERIOD)
	{
		FAULT("The period should be greater than %d uSec\n", MIN_PERIOD);
		return INVALID_PERIOD;
	}

	if(deadline_in_us < budget_in_us)
	{
		FAULT("The deadline should be at least as much as its budget\n");
		return INVALID_DEADLINE;
	}

	if(deadline_in_us > period_in_us)
	{
		FAULT("The deadline should be less than or equal to the period\n");
		return INVALID_DEADLINE;
	}


	if(budget_in_us < MIN_BUDGET)
	{
		FAULT("The budget should be greater than %d uSec\n", MIN_BUDGET);
		return INVALID_BUDGET;
	}

	// Conver the stack_size_in_bytes into number of words
	stack_size = stack_size_in_bytes >> 2; 

	task->type = PERIODIC;
	
	task->budget = budget_in_us;
	task->period = period_in_us;
	task->deadline = deadline_in_us;
	task->phase = phase_shift_in_us;
	task->stack = stack;
	task->stack_size = stack_size;
	task->top_of_stack = stack + stack_size;	// Stack grows bottom up
	task->task_function = periodic_entry_function;
	task->pdata = pdata;
	task->remaining_budget = 0;
	task->accumulated_budget = 0;
	task->stored_release_time = 0;
	task->exec_count = 0;
	task->TBE_count = 0;
	task->dline_miss_count = 0;
	//task->next_release_time = phase_shift_in_us;
	task->alarm_time = 0;

	OS_ENTER_CRITICAL();
	if(!ValidateNewThread(period_in_us, budget_in_us))
	{
		OS_EXIT_CRITICAL(); 
		FAULT("The total allocated CPU exceeds %d%", 100);
		return EXCEEDS_MAX_CPU;
	}

	task->id = ++g_task_id_counter;
	
	// Calculate the CPU usage of the current thread
	this_thread_cpu = CALC_THREAD_CPU_USAGE(period_in_us, budget_in_us);
	
	// Update the g_total_allocated_cpu
	g_total_allocated_cpu += this_thread_cpu;
	
	OS_EXIT_CRITICAL(); 	// Exit the critical section

	// Build a Stack for the new thread
	task->top_of_stack = C6713_BuildTaskStack(task->top_of_stack, 
		TaskEntryMain, task);

	OS_ENTER_CRITICAL();	// Enter critical section

	Nano_QueueInsert(&g_wait_q, task, phase_shift_in_us);		

	OS_EXIT_CRITICAL(); 	// Exit the critical section

	if(OS_IsRunning)
	{
		Nano_ReSchedule();
	}

  	return SUCCESS; 
}

///////////////////////////////////////////////////////////////////////////////
// Validation for sufficient CPU Budget
// ASSUMPTION: The interrupts are disabled when this function is invoked
///////////////////////////////////////////////////////////////////////////////
static BOOL ValidateNewThread(UINT32 period, UINT32 budget)
{
	// The current threads budget in terms of CPU usage %ge
	FP32 this_thread_cpu = CALC_THREAD_CPU_USAGE(period, budget);

	// Dont allow allocation of more than 90% of thr CPU
	if(g_total_allocated_cpu + this_thread_cpu >= 1.0)
	{
		return FALSE;
	}

	return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
// This is the main entry function for all periodic functions
///////////////////////////////////////////////////////////////////////////////
static void TaskEntryMain(void *pdata)
{
	OS_PeriodicTask * task = (OS_PeriodicTask *) pdata;

	while(1)
	{
		// Call the thread handler function
		task->task_function(task->pdata);

		OS_ENTER_CRITICAL();		
		task->exec_count++;

		// Suspend the current task. This task will be automatically 
		// woken up by the alarm.
		Nano_QueueDelete(&g_ready_q, task);
		if(task->alarm_time == g_next_wakeup_time)
		{
			// If the next wakeup time is same as the alarm time for the 
			// current task, just invalidate the next wakeup time so that
			// it is set again below
			g_next_wakeup_time = 0xFFFFFFFFFFFFFFFF;
		}
		if(task->deadline == task->period)
		{
			Nano_SetAlarm(task, task->alarm_time, FALSE);
		}
		else
		{
			Nano_SetAlarm(task, task->alarm_time + task->period - task->deadline, FALSE);
		}
		OS_EXIT_CRITICAL();

		Nano_ReSchedule();	// Now call reschedule function
	}
}

///////////////////////////////////////////////////////////////////////////////
// The task creation routine for Aperiodic tasks
///////////////////////////////////////////////////////////////////////////////
OS_Error OS_CreateAperiodicTask(UINT16 priority,
	UINT32 * stack, UINT32 stack_size_in_bytes,
	OS_AperiodicTask * task,
	void(* task_entry_function)(void * pdata),
	void * pdata)
{
	UINT32 stack_size;

	if(!task)
	{
		FAULT("Invalid Task");
		return INVALID_TASK;
	}

	if(!task_entry_function || !stack)
	{
		FAULT("One or more invalid %s arguments", "task");
		return INVALID_ARG;
	}

	if(stack_size_in_bytes < ONE_KB / 4) // Validate for minimum stack size 256 bytes
	{
		FAULT("Stack size should be at least %d bytes", ONE_KB / 4);
		return INSUFFICIENT_STACK;
	}

	// Conver the stack_size_in_bytes into number of words
	stack_size = stack_size_in_bytes >> 2; 

	task->stack = stack;
	task->stack_size = stack_size;
	task->top_of_stack = stack + stack_size;	// Stack grows bottom up
	task->top_of_stack = C6713_BuildTaskStack(task->top_of_stack, TaskEntryMain, task);
	task->task_function = task_entry_function;
	task->pdata = pdata;
	task->priority = 0;
	task->type = 0;
	task->id = 0;

	OS_ENTER_CRITICAL(); // Enter critical section
	Nano_QueueInsert(&g_ap_wait_q, task, 0); //add task to aperiodic wait queue
	OS_EXIT_CRITICAL(); // Exit the critical section

	return SUCCESS;
}

