///////////////////////////////////////////////////////////////////////////////
//	
//								Nano OS	Version	1.0
//						Copyright 2009 xxxxxxx, xxxxxxx
//	File:	Nano_Timer.c
//	Author: Balasubramanya Bhat
//	Description: Nano OS Timer implementation
//	
///////////////////////////////////////////////////////////////////////////////

#include "..\Nano_Types.h"
#include "Nano_Config.h"

// Note the way the following expression is put. We first do 
// 150 * us and then divide by 4. This is because if we first
// divide 150 by 4, there is a loss of precission. Since
// us will be generally big, in the below case the loss of
// precission is ignorable. 
// The same goes for the next line as well
#define CONVERT_us_TO_TICKS(us)		((CPU_CLOCK_IN_MHZ * us) / 4)
#define CONVERT_TICKS_TO_us(tick)	((tick * 4) / CPU_CLOCK_IN_MHZ)

#define T0_CTL	(volatile UINT32 *)0x01940000	//Timer Control Register of Timer-0
#define T0_PRD	(volatile UINT32 *)0x01940004	//Timer Period Register of Timer-0
#define T0_CNT	(volatile UINT32 *)0x01940008	//Timer Counter Register of Timer-0

#define T1_CTL	(volatile UINT32 *)0x01980000	//Timer Control Register of Timer-1
#define T1_PRD	(volatile UINT32 *)0x01980004	//Timer Period Register of Timer-1
#define T1_CNT	(volatile UINT32 *)0x01980008	//Timer Counter Register of Timer-1

extern cregister volatile unsigned int CSR ;
extern cregister volatile unsigned int IER ;
extern cregister volatile unsigned int ICR ;

UINT32 Nano_Timer1ISRHook();
UINT32 Nano_Timer2ISRHook();
UINT32 Nano_SetBudgetTimer(UINT32 delay_in_us);
void C6713_IntContextStore();
void ChangeToIntContext();

///////////////////////////////////////////////////////////////////////////////
// Function to initialize the timers 0 & 1
///////////////////////////////////////////////////////////////////////////////
void Nano_InitTimer ()
{
	static BOOL initialized = 0;

	if(!initialized)
	{
		// Place the timer in hold (HLD = 0, GO = 0). Note that after device
		// reset, the timer is already in the hold state.
		// Also clear the CP bit
		*T0_CTL &= ~0x1C0;
		*T1_CTL &= ~0x1C0;
		
		// Set the CLKSRC Bit
		*T0_CTL |= 0x200;
		*T1_CTL |= 0x200;

		// Initialize ICR, IER and CSR
		ICR = 0;
		IER |= 0xC000;

		// Set the initialized flag
		initialized = 1;
	}
}


///////////////////////////////////////////////////////////////////////////////
// Function to update the period in the timer
///////////////////////////////////////////////////////////////////////////////
BOOL Nano_UpdateTimer(UINT32 delay_in_us)
{
	UINT32 req_count = CONVERT_us_TO_TICKS(delay_in_us);
	UINT32 cur_count;

	// First Hold the Timer
	*T0_CTL &= ~0xC0;

	// Get the Terminal Value and the Current Counter
	cur_count = *T0_CNT;	 
	
	if(delay_in_us == 0)
	{
		*T0_CNT = 0; // The Timer is already in Hold State. Just reset the count to zero
	}
	else if(req_count > cur_count)
	{
		// The requested timeout is in the future. Update the terminal Count
		// and just resume counting
		*T0_PRD = req_count;
		*T0_CTL |= 0x80;
	}
	else
	{
		return 0; // The requested timeout has already passed or too close.
	}

	return 1;
}

///////////////////////////////////////////////////////////////////////////////
// Sets a new timeout for the budget timer 
///////////////////////////////////////////////////////////////////////////////
UINT32 Nano_SetBudgetTimer(UINT32 delay_in_us)
{
	UINT32 req_count = CONVERT_us_TO_TICKS(delay_in_us);
	UINT32 cur_count;

	// First Hold the Budget Timer
	*T1_CTL &= ~0xC0;

	// Get the Terminal Value and the Current Counter
	cur_count = CONVERT_TICKS_TO_us(*T1_CNT);	 
	
	if(delay_in_us == 0)
	{
		*T1_CNT = 0; // The Timer is already in Hold State. Just reset the count to zero
	}
	else 
	{
		// The requested timeout is in the future. Update the terminal Count
		// and just resume counting
		*T1_PRD = req_count;
		*T1_CTL |= 0xC0;
	}

	return cur_count;
}

///////////////////////////////////////////////////////////////////////////////
// Converts the current timer count into micro seconds and returns.
///////////////////////////////////////////////////////////////////////////////
UINT32 Nano_GetTime()
{
	return CONVERT_TICKS_TO_us(*T0_CNT);
}


///////////////////////////////////////////////////////////////////////////////
// Timer 1 ISR for Nano OS
///////////////////////////////////////////////////////////////////////////////
void OSTimer1ISR()
{
	// Offset the return address stored by OSTimer1ISR for calling below functions
	asm("	ADDK	8, SP");

	// Save the context of the current thread
	C6713_IntContextStore();

	// Change to Interrupt context
	ChangeToIntContext();

	// Call the ISR hook this might schedule a new thread
	Nano_Timer1ISRHook();	// Which context ??
}

///////////////////////////////////////////////////////////////////////////////
// Timer 1 ISR for Nano OS
///////////////////////////////////////////////////////////////////////////////
void OSTimer2ISR()
{
	// Offset the return address stored by OSTimer1ISR for calling below functions
	asm("	ADDK	8, SP");

	// Save the context of the current thread
	C6713_IntContextStore();

	// Change to Interrupt context
	ChangeToIntContext();

	// Call the ISR hook this might schedule a new thread
	Nano_Timer2ISRHook();	// Which context ??
}
